/*

Copyright (c) 2006, Eric Tiedemann
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:


* Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

      * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

      * Neither the name of Eric Tiedemann nor the names of his
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "rational.h"

namespace e7 {

  Rational::Rational(Bignum num, Bignum den) 
    : num(num), den(den) {
    canonicalize();
  }
  
  Rational::Rational(Bignum num) 
    : num(num), den(Bignum(int64_t(1))) {
    canonicalize();
  }
  
  Rational::Rational(int64_t num, int64_t den) 
    : num(num), den(den) {
    canonicalize();
  }

  Rational::Rational(int64_t num) 
    : num(num), den(Bignum(int64_t(1))) {
    canonicalize();
  }

  Rational::~Rational() {}

  void Rational::canonicalize() {
    // normalize to any negation being in the numerator
    if (den.negativep()) {
      num = -num;
      den = -den;
    }
    Bignum _gcd = gcd(num, den);
    num = quotient(num, _gcd);
    den = quotient(den, _gcd);
  }

  bool Rational::integralp() {
    return den == Bignum(int64_t(1));
  }

  const Rational operator+(Rational a, Rational b) {
    Bignum denlcm = lcm(a.den, b.den);
    return Rational(quotient(denlcm, a.den) * a.num
		    + quotient(denlcm, a.den) * a.num,
		    denlcm);
  }

  const Rational operator-(Rational a, Rational b) {
    Bignum denlcm = lcm(a.den, b.den);
    return Rational(quotient(denlcm, a.den) * a.num
		    - quotient(denlcm, a.den) * a.num,
		    denlcm);
  }
  
  const Rational operator*(Rational a, Rational b) {
    return Rational(a.num * b.num, a.den * b.den);
  }
  
  const Rational operator/(Rational a, Rational b) {
    return Rational(a.num * b.den, a.den * b.num);
  }
  
  
  bool operator==(Rational a, Rational b) {
    return bool(a.num == b.num && a.den == b.den);
  }
  
  bool operator!=(Rational a, Rational b) {
    return !bool(a.num == b.num && a.den == b.den);
  }
  
  bool Rational::cmp(Rational a, Rational b) {
    Bignum denlcm = lcm(a.den, b.den);
    Bignum a2 = a.num * quotient(denlcm, a.den);
    Bignum b2 = b.num * quotient(denlcm, b.den);
    if (a2 == b2) {
      return 0;
    } else if (a2 < b2) {
      return -1;
    } else {
      assert(a2 > b2);
      return 1;
    }
  }
  
  bool operator>(Rational a, Rational b) {
    return Rational::cmp(a, b) > 0;
  }
  
  bool operator<(Rational a, Rational b) {
    return Rational::cmp(a, b) < 0;
  }
  
  bool operator<=(Rational a, Rational b) {
    return Rational::cmp(a, b) <= 0;
  }
  
  bool operator>=(Rational a, Rational b) {
    return Rational::cmp(a, b) >= 0;
  }


std::ostream &operator<<(std::ostream &out, const Rational r) {
  out << r.num << "/" << r.den;
  return out;
}


  // FIXME factor with bignum expt()
  // FIXME do better :)
  Rational expt(Rational x, Bignum n) {
    Rational s = x, a(Bignum(int64_t(1)), Bignum(int64_t(1)));
    Bignum i = n;
    while (true) {
      if (i.oddp()) {
	a = a * s;
      }
      i = i.asl(-1);
      if (i.zerop()) {
	return a;
      }
      s = s * s;
    }
  }


}
